==> jview-0.1.2.tar.gz: jview-0.1.2/imageviewer.c <==
/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "prefs.h"
#include "imageviewer.h"

static void         image_viewer_class_init( ImageViewerClass *klass);
static void         image_viewer_init( ImageViewer *viewer);

static void         image_viewer_finalize( GtkObject *object);

static void         image_viewer_realize( GtkWidget *widget);
static void         image_viewer_size_request( GtkWidget *widget, GtkRequisition *requisition);
static gint         image_viewer_redraw( ImageViewer *viewer);
static gint         image_viewer_progress( Image *image, gfloat percentage, ImageViewer *viewer);
static void         image_viewer_draw( GtkWidget *widget, GdkRectangle *area);
static gint         image_viewer_expose( GtkWidget *widget, GdkEventExpose *event);

static GtkWidgetClass *parent_class;

static char           *pref_id = "jview/image_viewer";
static char           *pref_shrink_image = "shrink image";
static gint           shrink_image;

GtkType
image_viewer_get_type (void)
{
  static GtkType viewer_type = 0;

  if (!viewer_type)
    {
      GtkTypeInfo viewer_info =
      {
	"ImageViewer",
	sizeof (ImageViewer),
	sizeof (ImageViewerClass),
	(GtkClassInitFunc) image_viewer_class_init,
	(GtkObjectInitFunc) image_viewer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };

      viewer_type = gtk_type_unique (GTK_TYPE_WIDGET, &viewer_info);
    }

  return viewer_type;
}

static void         
image_viewer_class_init( ImageViewerClass *klass) {
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;
    PrefDef viewer_pref[] = {
	{ PREF_PUBLIC, pref_shrink_image,
	  "Shrink image to fit in window",
	  NULL, NULL,
	  GTK_TYPE_BOOL, "TRUE"}
    };

    pref_new_from_defs( viewer_pref, sizeof( viewer_pref) / sizeof( PrefDef), pref_id);
 
    object_class = (GtkObjectClass *) klass;
    widget_class = (GtkWidgetClass *) klass;

    parent_class = gtk_type_class( gtk_widget_get_type());

    object_class->finalize = image_viewer_finalize;

    widget_class->realize = image_viewer_realize;
    widget_class->size_request = image_viewer_size_request;
    widget_class->expose_event = image_viewer_expose;
    widget_class->draw   = image_viewer_draw;

    shrink_image = pref_get_bool( pref_shrink_image, pref_id);
}


static void
image_viewer_init( ImageViewer *viewer) {
    g_return_if_fail( viewer != NULL);

    GTK_WIDGET_UNSET_FLAGS( viewer, GTK_NO_WINDOW);

    viewer->image = NULL;
    viewer->new_image = NULL;
    viewer->drawing = FALSE;
    viewer->scale_factor = 1.0;
    viewer->need_redraw = FALSE;
}


static void
image_viewer_finalize( GtkObject *object) {
    ImageViewer *viewer;
    g_return_if_fail( object != NULL);

    viewer = IMAGE_VIEWER( object);

    if ( viewer->image){
	image_free( viewer->image);
	viewer->image = NULL;
    }

    if ( GTK_OBJECT_CLASS( parent_class)->finalize)
	( *GTK_OBJECT_CLASS( parent_class)->finalize) ( object);
}



GtkWidget*          
image_viewer_new( Image *image, ImageProgressFunc func, gpointer data) {
    ImageViewer *viewer;

    viewer = gtk_type_new( image_viewer_get_type());
    viewer->func = func;
    viewer->data = data;

    image_viewer_set_image( viewer, image);
    
    return GTK_WIDGET( viewer);
}

void                
image_viewer_set_image( ImageViewer *viewer, Image *image) {
    g_return_if_fail( viewer != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( viewer));

    
    if ( viewer->drawing) {
	viewer->new_image = image;
	viewer->need_redraw = TRUE;
    } else  {

	if ( viewer->image)
	    image_free( viewer->image);
	
	viewer->image = image ? image_copy( image) : NULL;
	if ( viewer->image)
	    image_free_preview( viewer->image);

	viewer->scale_factor = shrink_image ? -1.0 : 1.0;
	
	if ( GTK_WIDGET_VISIBLE( viewer))
	    gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);

	viewer->new_image = NULL;
    }
}

float
image_viewer_get_scale_factor( ImageViewer *viewer) {
    g_return_val_if_fail( viewer != NULL, 1.0);
    g_return_val_if_fail( IS_IMAGE_VIEWER( viewer), 1.0);

    if ( !viewer->image)
	return 1.0;

    return viewer->scale_factor < 0.0 ? 1.0 : viewer->scale_factor;
}

void
image_viewer_set_scale_factor( ImageViewer *viewer, float scale_factor) {
    g_return_if_fail( viewer != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( viewer));
    g_return_if_fail( scale_factor > 0.0 );

    viewer->scale_factor = scale_factor;
    if ( viewer->image) {
	if ( viewer->drawing) {
	    viewer->need_redraw = TRUE;
	    gtk_idle_add( (GtkFunction) image_viewer_redraw, viewer);
	} else {
	    image_free_pixmaps( viewer->image);
	    if ( GTK_WIDGET_VISIBLE( viewer))
		gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);
	}
    }
}

static void
image_viewer_realize( GtkWidget *widget) {
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( widget));

    GTK_WIDGET_SET_FLAGS( widget, GTK_REALIZED);

    /* widget window */
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual( widget);
    attributes.colormap = gtk_widget_get_colormap( widget);
    attributes.event_mask = (GDK_EXPOSURE_MASK |
			     GDK_BUTTON_PRESS_MASK |
			     GDK_BUTTON_RELEASE_MASK |
			     GDK_KEY_PRESS_MASK |
			     GDK_KEY_RELEASE_MASK);
    
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    
    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), 
				     &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, widget);

    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background( widget->style, widget->window, GTK_STATE_NORMAL);

}

static void         
image_viewer_size_request( GtkWidget *widget, GtkRequisition *requisition) {
    ImageViewer *viewer;

    viewer = IMAGE_VIEWER( widget);

    if ( viewer->image) {
	if ( viewer->scale_factor < 0.0) {
	    requisition->width = 10;
	    requisition->height = 10;

	} else {
	    requisition->width  = (int) (viewer->scale_factor * viewer->image->width);
	    requisition->height = (int) (viewer->scale_factor * viewer->image->height);
	}
    } 
}

static gint
image_viewer_redraw( ImageViewer *viewer) {
    if ( viewer->drawing)
	return TRUE;

    if ( viewer->new_image) /* called by image_viewer_progress */
	image_viewer_set_image( viewer, viewer->new_image);
    else if ( viewer->image) { /* called by image_set_scale_factor */
	image_free_pixmaps( viewer->image);
	if ( GTK_WIDGET_VISIBLE( viewer))
	    gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);
    }
    
    viewer->need_redraw = FALSE;

    return FALSE;
}

static gint
image_viewer_progress( Image *image, gfloat percentage, ImageViewer *viewer) {
    if ( viewer->need_redraw) {
	(*viewer->func) ( image, 1.0, viewer->data);
	gtk_idle_add( (GtkFunction) image_viewer_redraw, viewer);
	return FALSE;
    }

    if ( viewer->func)
	return (*viewer->func) ( image, percentage, viewer->data);

    return TRUE;
}


static void
image_viewer_draw( GtkWidget *widget, GdkRectangle *area) {
    ImageViewer  *viewer;
    GdkRectangle intersect_area;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    gint xsrc, ysrc, xdest, ydest;
    gint pixmap_width, pixmap_height;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( widget));

    viewer = IMAGE_VIEWER( widget);
    if ( viewer->drawing)
	return;

    viewer->drawing = TRUE;

    if ( GTK_WIDGET_DRAWABLE( widget)) {

	if ( !viewer->image) {
	    gdk_window_clear_area( widget->window,
			       area->x,
			       area->y,
			       area->width,
			       area->height);
	    viewer->drawing = FALSE;
	    return;
	}

	pixmap = image_get_pixmap( viewer->image);
	if ( pixmap) {
	    mask = image_get_mask( viewer->image);
	    
	    gdk_window_get_size( pixmap, &pixmap_width, &pixmap_height);
	    
	    xsrc = ysrc = 0;
	    xdest = ydest = 0;
	    if ( xdest < area->x) {
		xsrc = area->x - xdest;
		pixmap_width -= xsrc;
		xdest = area->x;
	    }
	    
	    if ( xdest +  pixmap_width > area->x + area->width)
		pixmap_width = (area->x + area->width) - xdest;
	
	    if ( ydest < area->y) {
		ysrc = area->y - ydest;
		pixmap_height -= ysrc;
		ydest = area->y;
	    }

	    if ( ydest + pixmap_height > area->y + area->height)
		pixmap_height = (area->y + area->height) - ydest;
	    
	    
	    if ( pixmap_width > 0 && pixmap_height > 0) {
		if ( mask) {
		    gdk_gc_set_clip_mask( widget->style->black_gc, mask);
		    gdk_gc_set_clip_origin( widget->style->black_gc, area->x, area->y);
		}
		
		
		gdk_draw_pixmap( widget->window,
				 widget->style->black_gc,
				 pixmap,
				 xsrc,
				 ysrc,
				 xdest, 
				 ydest,
				 pixmap_width, 
				 pixmap_height);
		
		
		if ( mask) {
		    gdk_gc_set_clip_mask( widget->style->black_gc, NULL);
		    gdk_gc_set_clip_origin( widget->style->black_gc, 0, 0);
		}
	    }

	} else {
	    gint flags;

	    xdest = ydest = 0;
	    flags = IMAGE_RENDER_PIXMAP | IMAGE_RENDER_IN_WINDOW | IMAGE_CACHE;
	    if ( viewer->scale_factor < 0.0) {
		pixmap_width  = widget->allocation.width;
		pixmap_height = widget->allocation.height;
		
   
	    } else {
		pixmap_width = viewer->scale_factor * viewer->image->width;
		pixmap_height = viewer->scale_factor * viewer->image->height;

		flags |= IMAGE_MAGNIFY_IF_SMALLER;
	    }
	    image_render( viewer->image,
			  widget->window,
			  widget->style->fg_gc[ GTK_STATE_NORMAL],
			  &widget->style->bg[ GTK_STATE_NORMAL],
			  xdest, ydest,
			  pixmap_width, 
			  pixmap_height,
			  flags, 
			  (ImageProgressFunc) image_viewer_progress, viewer);
	    
	    if ( viewer->scale_factor < 0.0 && viewer->image->preview) 
		viewer->scale_factor = (float) viewer->image->preview->width / viewer->image->width;

	}

	/* clear the part of the window that doesn't contain the pixmap */
	if ( area->x < xdest) 
	    gdk_window_clear_area( widget->window,
				   area->x,
				   area->y,
				   xdest - area->x,
				   area->height);
	if ( xdest + pixmap_width < area->x + area->width)
	    gdk_window_clear_area( widget->window,
				   xdest + pixmap_width,
				   area->y,
				   area->x + area->width - xdest - pixmap_width,
				   area->height);

	if ( area->y < ydest) 
	    gdk_window_clear_area( widget->window,
				   area->x,
				   area->y,
				   area->width,
				   ydest - area->y);
	if ( ydest + pixmap_height < area->y + area->height)
	    gdk_window_clear_area( widget->window,
				   area->x,
				   ydest + pixmap_height,
				   area->width,
				   area->y + area->height - ydest - pixmap_height);
				   

    }

    viewer->drawing = FALSE;
}


static gint  
image_viewer_expose( GtkWidget *widget, GdkEventExpose *event) {
    ImageViewer *viewer;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( event != NULL, FALSE);
    g_return_val_if_fail( IS_IMAGE_VIEWER( widget), FALSE);

    viewer = IMAGE_VIEWER( widget);
    if ( GTK_WIDGET_DRAWABLE( widget)) {
   
	if ( viewer->drawing) 
	    return TRUE;
	else if ( event->window == widget->window)
	    image_viewer_draw( widget, &event->area);

    }

    return FALSE;
    
}

==> jview-0.1.3.tar.gz: jview-0.1.3/imageviewer.c <==
/*  jview : Image viewer
 *  Copyright (C) 1998 david Ramboz
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "prefs.h"
#include "imageviewer.h"

static void         image_viewer_class_init( ImageViewerClass *klass);
static void         image_viewer_init( ImageViewer *viewer);

static void         image_viewer_finalize( GtkObject *object);

static void         image_viewer_realize( GtkWidget *widget);
static void         image_viewer_size_request( GtkWidget *widget, GtkRequisition *requisition);
static gint         image_viewer_redraw( ImageViewer *viewer);
static gint         image_viewer_progress( Image *image, gfloat percentage, ImageViewer *viewer);
static void         image_viewer_draw( GtkWidget *widget, GdkRectangle *area);
static gint         image_viewer_expose( GtkWidget *widget, GdkEventExpose *event);

static GtkWidgetClass *parent_class;

static char           *pref_id = "jview/image_viewer";
static char           *pref_shrink_image = "shrink image";


static gint           shrink_image;

GtkType
image_viewer_get_type (void)
{
  static GtkType viewer_type = 0;

  if (!viewer_type)
    {
      GtkTypeInfo viewer_info =
      {
	"ImageViewer",
	sizeof (ImageViewer),
	sizeof (ImageViewerClass),
	(GtkClassInitFunc) image_viewer_class_init,
	(GtkObjectInitFunc) image_viewer_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
	(GtkClassInitFunc) NULL,
      };

      viewer_type = gtk_type_unique (GTK_TYPE_WIDGET, &viewer_info);
    }

  return viewer_type;
}

static void         
image_viewer_class_init( ImageViewerClass *klass) {
    GtkObjectClass *object_class;
    GtkWidgetClass *widget_class;
    GtkArg *value;
    PrefDef viewer_pref[] = {
	{ PREF_PUBLIC, pref_shrink_image,
	  "Shrink image to fit in window",
	  NULL, NULL,
	  GTK_TYPE_BOOL, "TRUE"}
    };

    pref_new_from_defs( viewer_pref, sizeof( viewer_pref) / sizeof( PrefDef), pref_id);
 
    object_class = (GtkObjectClass *) klass;
    widget_class = (GtkWidgetClass *) klass;

    parent_class = gtk_type_class( gtk_widget_get_type());

    object_class->finalize = image_viewer_finalize;

    widget_class->realize = image_viewer_realize;
    widget_class->size_request = image_viewer_size_request;
    widget_class->expose_event = image_viewer_expose;
    widget_class->draw   = image_viewer_draw;

    value = pref_get_value( pref_shrink_image, pref_id);
    shrink_image = GTK_VALUE_BOOL( *value);
}


static void
image_viewer_init( ImageViewer *viewer) {
    g_return_if_fail( viewer != NULL);

    GTK_WIDGET_UNSET_FLAGS( viewer, GTK_NO_WINDOW);

    viewer->image = NULL;
    viewer->new_image = NULL;
    viewer->drawing = FALSE;
    viewer->scale_factor = 1.0;
    viewer->need_redraw = FALSE;
}


static void
image_viewer_finalize( GtkObject *object) {
    ImageViewer *viewer;
    g_return_if_fail( object != NULL);

    viewer = IMAGE_VIEWER( object);

    if ( viewer->image){
	image_free( viewer->image);
	viewer->image = NULL;
    }

    if ( GTK_OBJECT_CLASS( parent_class)->finalize)
	( *GTK_OBJECT_CLASS( parent_class)->finalize) ( object);
}



GtkWidget*          
image_viewer_new( Image *image, ImageProgressFunc func, gpointer data) {
    ImageViewer *viewer;

    viewer = gtk_type_new( image_viewer_get_type());
    viewer->func = func;
    viewer->data = data;

    image_viewer_set_image( viewer, image);
    
    return GTK_WIDGET( viewer);
}

void                
image_viewer_set_image( ImageViewer *viewer, Image *image) {
    g_return_if_fail( viewer != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( viewer));

    
    if ( viewer->drawing) {
	viewer->new_image = image;
	viewer->need_redraw = TRUE;
    } else  {

	if ( viewer->image) {
	    image_free( viewer->image);
	    viewer->image = NULL;
	}

	if ( image && !image_read_header( image))
	    return;

	viewer->image = image ? image_copy( image) : NULL;

	viewer->scale_factor = shrink_image ? -1.0 : 1.0;
	
	if ( GTK_WIDGET_VISIBLE( viewer))
	    gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);

	viewer->new_image = NULL;
    }
}

float
image_viewer_get_scale_factor( ImageViewer *viewer) {
    g_return_val_if_fail( viewer != NULL, 1.0);
    g_return_val_if_fail( IS_IMAGE_VIEWER( viewer), 1.0);

    if ( !viewer->image)
	return 1.0;

    return viewer->scale_factor < 0.0 ? 1.0 : viewer->scale_factor;
}

void
image_viewer_set_scale_factor( ImageViewer *viewer, float scale_factor) {
    g_return_if_fail( viewer != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( viewer));
    g_return_if_fail( scale_factor > 0.0 );

    viewer->scale_factor = scale_factor;
    if ( viewer->image) {
	if ( viewer->drawing) {
	    viewer->need_redraw = TRUE;
	    gtk_idle_add( (GtkFunction) image_viewer_redraw, viewer);
	} else {
	    image_free_pixmaps( viewer->image);
	    if ( GTK_WIDGET_VISIBLE( viewer))
		gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);
	}
    }
}

static void
image_viewer_realize( GtkWidget *widget) {
    GdkWindowAttr attributes;
    gint attributes_mask;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( widget));

    GTK_WIDGET_SET_FLAGS( widget, GTK_REALIZED);

    /* widget window */
    attributes.window_type = GDK_WINDOW_CHILD;
    attributes.x = widget->allocation.x;
    attributes.y = widget->allocation.y;
    attributes.width = widget->allocation.width;
    attributes.height = widget->allocation.height;
    attributes.wclass = GDK_INPUT_OUTPUT;
    attributes.visual = gtk_widget_get_visual( widget);
    attributes.colormap = gtk_widget_get_colormap( widget);
    attributes.event_mask = (GDK_EXPOSURE_MASK |
			     GDK_BUTTON_PRESS_MASK |
			     GDK_BUTTON_RELEASE_MASK |
			     GDK_KEY_PRESS_MASK |
			     GDK_KEY_RELEASE_MASK);
    
    attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
    
    widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), 
				     &attributes, attributes_mask);
    gdk_window_set_user_data (widget->window, widget);

    widget->style = gtk_style_attach (widget->style, widget->window);
    gtk_style_set_background( widget->style, widget->window, GTK_STATE_NORMAL);

}

static void         
image_viewer_size_request( GtkWidget *widget, GtkRequisition *requisition) {
    ImageViewer *viewer;

    viewer = IMAGE_VIEWER( widget);

    if ( viewer->image) {
	if ( viewer->scale_factor < 0.0) {
	    requisition->width = 10;
	    requisition->height = 10;

	} else {
	    requisition->width  = (int) (viewer->scale_factor * viewer->image->width);
	    requisition->height = (int) (viewer->scale_factor * viewer->image->height);
	}
    } 
}

static gint
image_viewer_redraw( ImageViewer *viewer) {
    if ( viewer->drawing)
	return TRUE;

    if ( viewer->new_image) /* called by image_viewer_progress */
	image_viewer_set_image( viewer, viewer->new_image);
    else if ( viewer->image) { /* called by image_set_scale_factor */
	image_free_pixmaps( viewer->image);
	if ( GTK_WIDGET_VISIBLE( viewer))
	    gtk_widget_queue_resize( GTK_WIDGET( viewer)->parent);
    }
    
    viewer->need_redraw = FALSE;

    return FALSE;
}

static gint
image_viewer_progress( Image *image, gfloat percentage, ImageViewer *viewer) {
    if ( viewer->need_redraw) {
	(*viewer->func) ( image, 1.0, viewer->data);
	gtk_idle_add( (GtkFunction) image_viewer_redraw, viewer);
	return FALSE;
    }

    if ( viewer->func)
	return (*viewer->func) ( image, percentage, viewer->data);

    return TRUE;
}


static void
image_viewer_draw( GtkWidget *widget, GdkRectangle *area) {
    ImageViewer  *viewer;
    GdkPixmap *pixmap;
    GdkBitmap *mask;
    gint xsrc, ysrc, xdest, ydest;
    gint pixmap_width, pixmap_height;

    g_return_if_fail( widget != NULL);
    g_return_if_fail( IS_IMAGE_VIEWER( widget));

    viewer = IMAGE_VIEWER( widget);
    if ( viewer->drawing)
	return;

    viewer->drawing = TRUE;

    if ( GTK_WIDGET_DRAWABLE( widget)) {

	gdk_window_clear_area( widget->window,
			       area->x,
			       area->y,
			       area->width,
			       area->height);

	if ( !viewer->image)
	    return;

	pixmap = image_get_pixmap( viewer->image);
	if ( pixmap) {
	    mask = image_get_mask( viewer->image);
	    
	    gdk_window_get_size( pixmap, &pixmap_width, &pixmap_height);
	    
	    xsrc = ysrc = 0;
	    xdest = widget->allocation.width > pixmap_width ? (widget->allocation.width - pixmap_width) / 2 : 0;
	    ydest = widget->allocation.height > pixmap_height ? (widget->allocation.height - pixmap_height) / 2 : 0;
	    if ( xdest < area->x) {
		xsrc = area->x - xdest;
		pixmap_width -= xsrc;
		xdest = area->x;
	    }
	    
	    if ( xdest +  pixmap_width > area->x + area->width)
		pixmap_width = (area->x + area->width) - xdest;
	
	    if ( ydest < area->y) {
		ysrc = area->y - ydest;
		pixmap_height -= ysrc;
		ydest = area->y;
	    }

	    if ( ydest + pixmap_height > area->y + area->height)
		pixmap_height = (area->y + area->height) - ydest;
	    
	    
	    if ( pixmap_width > 0 && pixmap_height > 0) {
		if ( mask) {
		    gdk_gc_set_clip_mask( widget->style->black_gc, mask);
		    gdk_gc_set_clip_origin( widget->style->black_gc, area->x, area->y);
		}
		
		
		gdk_draw_pixmap( widget->window,
				 widget->style->black_gc,
				 pixmap,
				 xsrc,
				 ysrc,
				 xdest, 
				 ydest,
				 pixmap_width, 
				 pixmap_height);
		
		
		if ( mask) {
		    gdk_gc_set_clip_mask( widget->style->black_gc, NULL);
		    gdk_gc_set_clip_origin( widget->style->black_gc, 0, 0);
		}
	    }

	} else {
	    gint flags;

	    flags = IMAGE_RENDER_PIXMAP | IMAGE_RENDER_IN_WINDOW | IMAGE_CACHE;
	    if ( viewer->scale_factor < 0.0) {

		image_get_preview_size( viewer->image,
					MIN( viewer->image->width, widget->allocation.width),
					MIN( viewer->image->height, widget->allocation.height),
					flags,
					&pixmap_width,
					&pixmap_height);
   
	    } else {
		flags |= IMAGE_MAGNIFY_IF_SMALLER;

		image_get_preview_size( viewer->image,
					viewer->scale_factor * viewer->image->width,
					viewer->scale_factor * viewer->image->height,
					flags,
					&pixmap_width,
					&pixmap_height);
   

	    }
	    xdest = widget->allocation.width > pixmap_width ? (widget->allocation.width - pixmap_width) / 2 : 0;
	    ydest = widget->allocation.height > pixmap_height ? (widget->allocation.height - pixmap_height) / 2 : 0;
	    image_render( viewer->image,
			  widget->window,
			  widget->style->fg_gc[ GTK_STATE_NORMAL],
			  &widget->style->bg[ GTK_STATE_NORMAL],
			  xdest, ydest,
			  pixmap_width, 
			  pixmap_height,
			  flags, 
			  (ImageProgressFunc) image_viewer_progress, viewer);
	    
	    if ( viewer->scale_factor < 0.0 && viewer->image->preview) 
		viewer->scale_factor = (float) viewer->image->preview->width / viewer->image->width;

	}

	/* clear the part of the window that doesn't contain the pixmap */
	if ( area->x < xdest) 
	    gdk_window_clear_area( widget->window,
				   area->x,
				   area->y,
				   xdest - area->x,
				   area->height);
	if ( xdest + pixmap_width < area->x + area->width)
	    gdk_window_clear_area( widget->window,
				   xdest + pixmap_width,
				   area->y,
				   area->x + area->width - xdest - pixmap_width,
				   area->height);

	if ( area->y < ydest) 
	    gdk_window_clear_area( widget->window,
				   area->x,
				   area->y,
				   area->width,
				   ydest - area->y);
	if ( ydest + pixmap_height < area->y + area->height)
	    gdk_window_clear_area( widget->window,
				   area->x,
				   ydest + pixmap_height,
				   area->width,
				   area->y + area->height - ydest - pixmap_height);
				   

    }

    viewer->drawing = FALSE;
}


static gint  
image_viewer_expose( GtkWidget *widget, GdkEventExpose *event) {
    ImageViewer *viewer;

    g_return_val_if_fail( widget != NULL, FALSE);
    g_return_val_if_fail( event != NULL, FALSE);
    g_return_val_if_fail( IS_IMAGE_VIEWER( widget), FALSE);

    viewer = IMAGE_VIEWER( widget);
    if ( GTK_WIDGET_DRAWABLE( widget)) {
   
	if ( viewer->drawing) 
	    return TRUE;
	else if ( event->window == widget->window)
	    image_viewer_draw( widget, &event->area);

    }

    return FALSE;
    
}
